#include	<gsl/gsl_rng.h>
#include	<iostream>
#include	<stdint.h>
#include	<cstdlib>
#include	<cstring>
#include	<cfloat>
#include	<cstdio>
#include	<vector>
#include	<cmath>
#include	<ctime>
#define	WordShift	6
#define	WordMod	63
#define	MaskHA	0x80
#define	MaskHB	0x40
#define	MaskG	0x3f
using	namespace	std;
typedef	float	fast;
typedef	double	real;
typedef	long	double	high;
typedef	unsigned	uint;
typedef uint64_t	word;
const	fast	norm=powf(FLT_MIN,	2.0f/3.0f);

class	Impute{
private:
	gsl_rng	*rng;
	uint	hn,	pn,	en,	wn;
	vector<word>	haps,	hnew;
	vector<uint>	hsum;
	vector<fast>	tran,	emit;
	
	fast	pc[4][4];	
	void	set1(word	*P,	uint	I){	P[I>>WordShift]|=static_cast<word>(1)<<(I&WordMod);	}
	void	set0(word	*P,	uint	I){	P[I>>WordShift]&=~(static_cast<word>(1)<<(I&WordMod));	}
	word	test(word	*P,	uint	I){	return	(P[I>>WordShift]>>(I&WordMod))&static_cast<word>(1);	}
	fast	hmm_like(uint	I,	uint	*P);
	fast	hmm(uint	I,	uint	*P,	fast	S);
	void	hmm_work(uint	I,	uint	*P,	fast	S);
	fast	solve(uint	I,	uint	N,	fast	S);
	void	replace(uint	I);
	void	result(void);
public:
	uint	in,	mn,	bn,	sn;
	vector<real>	posi;
	vector<fast>	prob;
	vector<char>	geno;
	
	Impute(){	rng=gsl_rng_alloc(gsl_rng_default);	gsl_rng_set(rng,	time(NULL));	}
	~Impute(){	gsl_rng_free(rng);	}
	bool	load_geno(const	char	*F,	fast	M=0);
	void	initialize(real	D);
	void	estimate(void);
	void	save_haps(const	char	*F);
};

bool	Impute::load_geno(const	char	*F,	fast	M){
	FILE	*f=fopen(F,	"rt");
	if(f==NULL)	return	false;
	mn=in=0;	prob.clear();	posi.clear();
	char	*buff=new	char[1024*1024];	
	const	fast	miss=1.0/3;
	while(fscanf(f,	"%s",	buff)!=EOF)	if(*buff){
		in=0;	mn++;
		for(char	*p=buff;	*p;	p++,	in++){
			switch(*p){
			case	'1':	prob.push_back(2*M*(1-M));	prob.push_back(M*M);	break;
			case	'2':	prob.push_back(1-2*M*(1-M));	prob.push_back(M*(1-M));	break;
			case	'3':	prob.push_back(2*M*(1-M));	prob.push_back((1-M)*(1-M));	break;
			default:	prob.push_back(miss);	prob.push_back(miss);	break;
			}
		}
	}
	delete	[]	buff;
	fclose(f);
	if(2*in*mn!=prob.size())	return	false;
	return	true;
}

void	Impute::initialize(real	D){
	wn=(mn&WordMod)?(mn>>WordShift)+1:(mn>>WordShift);
	hn=in*2;	haps.resize(hn*wn);	hnew.resize(hn*wn);
	hsum.resize(hn*mn);
	pn=3*mn;	tran.resize(pn);	vector<fast>	temp(in*pn);	
	en=4*mn;	emit.resize(in*en);	
	
	fast	mu=0,	rho;
	for(uint	i=1;	i<hn;	i++)	mu+=1.0/i;
	if(posi.size()!=mn){	posi.resize(mn);	for(uint	m=0;	m<mn;	m++)	posi[m]=m;	}
	mu=1/mu;	rho=0.5*mu*(mn-1)/(posi[mn-1]-posi[0])/D;	mu=mu/(hn+mu);
	for(uint	m=mn-1;	m;	m--){
		posi[m]=(posi[m]-posi[m-1])*rho;	fast	r=posi[m]/(posi[m]+hn);
		tran[m*3]=(1-r)*(1-r);	tran[m*3+1]=r*(1-r);	tran[m*3+2]=r*r;
	}
	pc[0][0]=pc[1][1]=pc[2][2]=pc[3][3]=(1-mu)*(1-mu);	
	pc[0][1]=pc[0][2]=pc[1][0]=pc[1][3]=pc[2][0]=pc[2][3]=pc[3][1]=pc[3][2]=mu*(1-mu);
	pc[0][3]=pc[1][2]=pc[2][1]=pc[3][0]=mu*mu;
	
	for(uint	i=0;	i<in;	i++){
		word	*ha=&haps[i*2*wn],	*hb=ha+wn;
		fast	*t=&temp[i*pn],	*e=&emit[i*en],	*p=&prob[i*2];
		for(uint	m=0;	m<mn;	m++,	t+=3,	e+=4,	p+=hn){
			if(gsl_rng_get(rng)&1)	set1(ha,	m);
			if(gsl_rng_get(rng)&1)	set1(hb,	m);
			t[0]=fmaxf(1-p[0]-p[1],0.0f);	t[1]=fmaxf(p[0],0.0f);	t[2]=fmaxf(p[1],0.0f);
			for(uint	j=0;	j<4;	j++)
				e[j]=pc[j][0]*t[0]+pc[j][1]*t[1]+pc[j][2]*t[1]+pc[j][3]*t[2];
			
		}
	}
	swap(temp,	prob);
}

fast	Impute::hmm_like(uint	I,	uint	*P){
	word	*f0=&haps[P[0]*wn],	*f1=&haps[P[1]*wn],	*m0=&haps[P[2]*wn],	*m1=&haps[P[3]*wn];	
	fast	*e=&emit[I*en],	*t=&tran[0],	sum,	score=0;
	fast	l00=0.25f*e[(test(f0,0)<<1)|test(m0,0)],	l01=0.25f*e[(test(f0,0)<<1)|test(m1,0)];
	fast	l10=0.25f*e[(test(f1,0)<<1)|test(m0,0)],	l11=0.25f*e[(test(f1,0)<<1)|test(m1,0)];
	fast	b00,	b01,	b10,	b11;
	e+=4;	t+=3;
	for(uint	m=1;	m<mn;	m++,	e+=4,	t+=3){
		b00=l00*t[0]+(l01+l10)*t[1]+l11*t[2];	b01=l01*t[0]+(l00+l11)*t[1]+l10*t[2];
		b10=l10*t[0]+(l00+l11)*t[1]+l01*t[2];	b11=l11*t[0]+(l01+l10)*t[1]+l00*t[2];
		l00=b00*e[(test(f0,m)<<1)|test(m0,m)];	l01=b01*e[(test(f0,m)<<1)|test(m1,m)];
		l10=b10*e[(test(f1,m)<<1)|test(m0,m)];	l11=b11*e[(test(f1,m)<<1)|test(m1,m)];
		if((sum=l00+l01+l10+l11)<norm){
			sum=1.0f/sum;	score-=logf(sum);
			l00*=sum;	l01*=sum;	l10*=sum;	l11*=sum;
		}
	}
	return	score+logf(l00+l01+l10+l11);
}

void	Impute::hmm_work(uint	I,	uint	*P,	fast	S){
	word	*f0=&haps[P[0]*wn],	*f1=&haps[P[1]*wn],	*m0=&haps[P[2]*wn],	*m1=&haps[P[3]*wn];	
	//	backward recursion
	vector<fast>	beta(mn*4);
	fast	*e=&emit[(I+1)*en-4],	*t=&tran[(mn-1)*3],	sum,	*b=&beta[(mn-1)*4];
	fast	l00=0,	l01=0,	l10=0,	l11=0;
	fast	b00,	b01,	b10,	b11;
	
	b[0]=b[1]=b[2]=b[3]=1;
	for(uint	m=mn-1;	m;	m--,	e-=4,	t-=3){
		b00=b[0]*e[(test(f0,m)<<1)|test(m0,m)];	b01=b[1]*e[(test(f0,m)<<1)|test(m1,m)];
		b10=b[2]*e[(test(f1,m)<<1)|test(m0,m)];	b11=b[3]*e[(test(f1,m)<<1)|test(m1,m)];
		b-=4;
		b[0]=b00*t[0]+(b01+b10)*t[1]+b11*t[2];	b[1]=b01*t[0]+(b00+b11)*t[1]+b10*t[2];
		b[2]=b10*t[0]+(b00+b11)*t[1]+b01*t[2];	b[3]=b11*t[0]+(b01+b10)*t[1]+b00*t[2];
		sum=1.0f/(b[0]+b[1]+b[2]+b[3]);	b[0]*=sum;	b[1]*=sum;	b[2]*=sum;	b[3]*=sum;
	}
	
	//	forward sampling	
	word	*ha=&hnew[I*2*wn],	*hb=ha+wn;
	fast	*p=&prob[I*pn];	e=&emit[I*en];	t=&tran[0];	b=&beta[0];
	uint	s00,	s01,	s10,	s11;
	for(uint	m=0;	m<mn;	m++,	e+=4,	t+=3,	p+=3,	b+=4){
		s00=(test(f0,m)<<1)|test(m0,m);	s01=(test(f0,m)<<1)|test(m1,m);
		s10=(test(f1,m)<<1)|test(m0,m);	s11=(test(f1,m)<<1)|test(m1,m);
		if(m){
			b00=l00*t[0]+l01*t[1]+l10*t[1]+l11*t[2],	b01=l00*t[1]+l01*t[0]+l10*t[2]+l11*t[1];
			b10=l00*t[1]+l01*t[2]+l10*t[0]+l11*t[1],	b11=l00*t[2]+l01*t[1]+l10*t[1]+l11*t[0];
			l00=b00*e[s00];	l01=b01*e[s01];	l10=b10*e[s10];	l11=b11*e[s11];
		}
		else{	l00=0.25f*e[s00];	l01=0.25f*e[s01];	l10=0.25f*e[s10];	l11=0.25f*e[s11];	}	
		sum=1.0f/(l00+l01+l10+l11);	l00*=sum;	l01*=sum;	l10*=sum;	l11*=sum;
		fast	p00=l00*b[0],	p01=l01*b[1],	p10=l10*b[2],	p11=l11*b[3];
		fast	c00=powf(p[0]*(p00*pc[s00][0]+p01*pc[s01][0]+p10*pc[s10][0]+p11*pc[s11][0]),	S);
		fast	c01=powf(p[1]*(p00*pc[s00][1]+p01*pc[s01][1]+p10*pc[s10][1]+p11*pc[s11][1]),	S);
		fast	c10=powf(p[1]*(p00*pc[s00][2]+p01*pc[s01][2]+p10*pc[s10][2]+p11*pc[s11][2]),	S);
		fast	c11=powf(p[2]*(p00*pc[s00][3]+p01*pc[s01][3]+p10*pc[s10][3]+p11*pc[s11][3]),	S);
		sum=gsl_rng_uniform(rng)*(c00+c01+c10+c11);
		if(sum<c00){	set0(ha,m);	set0(hb,m);	}
		else	if(sum<c00+c01){	set0(ha,m);	set1(hb,m);	}
		else	if(sum<c00+c01+c10){	set1(ha,m);	set0(hb,m);	}
		else{	set1(ha,m);	set1(hb,m);	}
	}
}

fast	Impute::solve(uint	I,	uint	N,	fast	S){
	uint	p[4];
	for(uint	j=0;	j<4;	j++){	
		do	p[j]=gsl_rng_get(rng)%hn;	while(p[j]/2==I);
	}
	fast	curr=hmm_like(I,	p);
		
	for(uint	n=0;	n<N;	n++){
		uint	rp=gsl_rng_get(rng)&3,	oh=p[rp];
		do	p[rp]=gsl_rng_get(rng)%hn;	while(p[rp]/2==I);
		fast	prop=hmm_like(I,	p);
		if(prop>curr||gsl_rng_uniform(rng)<expf((prop-curr)*S))	curr=prop;
		else	p[rp]=oh;
	}
	hmm_work(I,	p,	S);
	return	curr;	
}

void	Impute::estimate(void){
	for(uint	n=0;	n<bn+sn;	n++){
		fast	sum=0,	pen=fminf(2*(n+1.0f)/bn,	1);	pen*=pen;
		#pragma omp parallel for
		for(uint	i=0;	i<in;	i++)	sum+=solve(i,	1024,	pen);
		swap(hnew,	haps);
		if(n>=bn){
			#pragma omp parallel for
			for(uint	i=0;	i<in;	i++)	replace(i);
		}
		fprintf(stderr,	"%u\t%.3f\t%.3f\r",	n,	pen,	sum/in/mn);
	}
	fputc('\n',	stderr);
	result();	
}

void	Impute::replace(uint	I){
	word	*oa=&haps[I*2*wn],	*ob=oa+wn;
	uint	*ha=&hsum[I*2*mn],	*hb=ha+mn;
	uint	sis=0,	tra=0;
	for(uint	m=0;	m<mn;	m++){	
		if(test(oa,m)){	sis+=ha[m];	tra+=hb[m];	}
		if(test(ob,m)){	sis+=hb[m];	tra+=ha[m];	}
	}
	if(sis>tra)	for(uint	m=0;	m<mn;	m++){	ha[m]+=test(oa,m);	hb[m]+=test(ob,m);	}
	else	for(uint	m=0;	m<mn;	m++){	ha[m]+=test(ob,m);	hb[m]+=test(oa,m);	}	
}

void	Impute::result(void){
	prob.resize(mn*hn);	geno.resize(mn*in);
	fast	norm=1.0/sn;
	for(uint	m=0;	m<mn;	m++){
		fast	*p=&prob[m*hn];	char	*g=&geno[m*in];
		for(uint	i=0;	i<in;	i++,	g++,	p+=2){
			p[0]=hsum[i*2*mn+m]*norm;	p[1]=hsum[(i*2+1)*mn+m]*norm;
			fast	prr=(1-p[0])*(1-p[1]),	pra=(1-p[0])*p[1]+p[0]*(1-p[1]),	paa=p[0]*p[1];
			if(prr>=pra&&prr>=paa)	*g=0;
			else	if(pra>=prr&&pra>=paa)	*g=1;
			else	*g=2;
			posi[m]+=*g;
			if(p[0]>0.5){	*g|=MaskHA;	set1(&haps[i*2*wn],	m);	}	else	set0(&haps[i*2*wn],	m);
			if(p[1]>0.5){	*g|=MaskHB;	set1(&haps[(i*2+1)*wn],	m);	}	else	set0(&haps[(i*2+1)*wn],	m);
		}
	}
}

void	Impute::save_haps(const	char	*F){
	FILE	*f=fopen(F,	"wt");

	for(uint	m=0;	m<mn;	m++){
		char	*g=&geno[m*in];
		for(uint	i=0;	i<in;	i++,	g++){
			fputc((*g&MaskHA)?'1':'0',	f);
			fputc((*g&MaskHB)?'1':'0',	f);
		}
		fputc('\n',	f);
	}
	fclose(f);
}
/*
int	main(int	ac,	char	**av){
	if(ac<5)	return	0;
	Impute	lp;	lp.bn=atoi(av[4]);	lp.sn=lp.bn;
	lp.load_geno(av[1],	0);
	lp.initialize(atof(av[3]));
	lp.estimate();
	lp.save_haps(av[2]);
	return	0;
}
*/
