#include	<stdint.h>
#include	<cstdlib>
#include	<cstdio>
#include	<string>
#include	<cmath>
#include	<sam.h>
#include	<map>
using	namespace	std;
typedef	const	char	cchr;
typedef	unsigned	uint;
typedef	float	fast;

class	Quality{
private:
	struct	Group{
		uint64_t	mis[94],	tot[94];
		fast	qua[94];
		Group(){	
			for(uint	i=0;	i<94;	i++)	qua[i]=mis[i]=tot[i]=0;
		}
	};
	
	uint	read_info(bam1_t	*B,	string	&S);
public:
	static	fast	phred[256];
	map<string,	Group>	group;
	bool	null;	//	no model available	
	cchr	*refs;	//	reference sequence
	uint	refl;	//	length of the reference sequence
	uint	mask;	//	mask filter for FLAG filed in SAM format
	fast	oper;	//	operation rate filter: oper=#edit/read_length

	Quality(){	for(uint	i=0;	i<256;	i++)	phred[i]=1-exp10f(-0.1*i);	}
	void	push(bam1_t	*B);
	void	save(cchr	*F);
	bool	load(cchr	*F);
	void	adjust(bam1_t	*B);
};

fast	Quality::phred[256];

uint	Quality::read_info(bam1_t	*B,	string	&S){
	uint32_t	opers=0,	pseq=0,	pref=B->core.pos,	cn=B->core.n_cigar,	*cp=bam1_cigar(B);
	uint8_t	*seq=bam1_seq(B),	*qua=bam1_qual(B),	*s=bam1_aux(B);
	while(s<B->data+B->data_len){
		uint8_t	type,	key[2];
		key[0]=s[0];	key[1]=s[1];
		s+=2;	type=*s;	s++;
		if(key[0]=='C'&&key[1]=='M')	opers=*s;
		else	if(key[0]=='R'&&key[1]=='G')	S=(char*)s;
		else	if(key[0]=='O'&&key[1]=='Q'){
			for(int	i=0;	i<B->core.l_qseq;	i++)	qua[i]=s[i]-33;
		}
		if (type=='A'||type =='C'||type == 'c')	s++;
		else if (type == 'S'||type == 's')	s+=2;
		else if (type == 'I'||type == 'i'||type == 'f')	s+=4;
		else if (type == 'd')	s+=8;
		else if (type == 'Z'||type=='H'){	while(*s)	s++; s++;	}
	}
	for(uint	c=0;	c<cn;	c++){
		uint	clen=(cp[c]>>BAM_CIGAR_SHIFT);
		switch(cp[c]&BAM_CIGAR_MASK){
		case	BAM_CMATCH:
			for(uint	i=0;	i<clen;	i++,	pseq++,	pref++){
				char	cb=bam_nt16_rev_table[bam1_seqi(seq,	pseq)];
				char	rb=pref<refl?toupper(refs[pref]):'N';
				if(cb!='='&&rb!='N'&&cb!=rb)	opers++;
			}
			break;
		case	BAM_CINS:	case	BAM_CSOFT_CLIP:	pseq+=clen;	opers++;	break;
		case	BAM_CDEL:	case	BAM_CREF_SKIP:	pref+=clen;	opers++;	break;
		}
	}	
	return	opers;
}
	
void	Quality::push(bam1_t	*B){
	string	rg="default";
	if(B->core.flag&mask
	||!B->core.n_cigar
	||!B->core.qual
	||read_info(B,	rg)>oper*B->core.l_qseq
	)	return;
	uint32_t	pseq=0,	pref=B->core.pos,	cn=B->core.n_cigar,	*cp=bam1_cigar(B);
	uint8_t	*seq=bam1_seq(B),	*qua=bam1_qual(B);
	Group	&g=group[rg];
	for(uint	c=0;	c<cn;	c++){
		uint	clen=(cp[c]>>BAM_CIGAR_SHIFT);
		switch(cp[c]&BAM_CIGAR_MASK){
		case	BAM_CMATCH:
			for(uint	i=0;	i<clen;	i++,	pseq++,	pref++){
				char	cb=bam_nt16_rev_table[bam1_seqi(seq,	pseq)];
				char	rb=pref<refl?toupper(refs[pref]):'N';
				g.tot[qua[pseq]]++;	
				if(cb!='='&&rb!='N'&&cb!=rb)	g.mis[qua[pseq]]++;
			}
			break;
		case	BAM_CINS:	case	BAM_CSOFT_CLIP:	pseq+=clen;	break;
		case	BAM_CDEL:	case	BAM_CREF_SKIP:	pref+=clen;	break;
		}
	}	
}

void	Quality::adjust(bam1_t	*B){
	string	rg="default";
	if(null){	memset(bam1_qual(B),	255,	B->core.l_qseq);	return;	}
	if(B->core.flag&mask
	||!B->core.n_cigar
	||!B->core.qual
	||read_info(B,	rg)>oper*B->core.l_qseq
	){	memset(bam1_qual(B),	0,	B->core.l_qseq);	return;	}

	fast	*q=group[rg].qua,	mapq=phred[B->core.qual];
	uint8_t	*qua=bam1_qual(B);

	for(int	i=0;	i<B->core.l_qseq;	i++)	qua[i]=(uint8_t)(mapq*q[qua[i]]*255+0.5);
}

void	Quality::save(cchr	*F){
	string	s=F;	s+=".qc";
	FILE	*f=fopen(s.c_str(),	"wt");
	if(f==NULL)	return;
	
	double	mis=0,	tot=0;
	for(map<string,	Group>::iterator	mi=group.begin();	mi!=group.end();	++mi){
		for(uint	i=0;	i<94;	i++)	if(mi->second.tot[i]){
			double	p=1-(double)mi->second.mis[i]/mi->second.tot[i];
			mis+=p*mi->second.mis[i];	tot+=p*mi->second.tot[i];
		}
	}
	fprintf(f,	"mask\t%u\noper\t%g\n",	mask,	oper);
	fprintf(f,	"%lu",	group.size());
	for(map<string,	Group>::iterator	mi=group.begin();	mi!=group.end();	++mi)
		fprintf(f,	"\t%s",	mi->first.c_str());
	
	for(uint	i=0;	i<94;	i++){
		fprintf(f,	"\n%u",	i);
		for(map<string,	Group>::iterator	mi=group.begin();	mi!=group.end();	++mi){
			fprintf(f,	"\t%.3f",	mi->second.tot[i]?100.0*mi->second.mis[i]/mi->second.tot[i]:0);
		}
	}
	fclose(f);	
}

bool	Quality::load(cchr	*F){
	null=true;
	string	s=F;	s+=".qc";
	FILE	*f=fopen(s.c_str(),	"rt");
	if(f==NULL)	return	false;
	uint	ng;
	if(fscanf(f,	"%*s%u%*s%f%u",	&mask,	&oper,	&ng)==EOF)	return	false;
	group.clear();
	for(uint	i=0;	i<ng;	i++){
		char	rg[1024];
		if(fscanf(f,	"%s",	rg)==EOF)	return	false;
		group[rg];
	}
	for(uint	i=0;	i<94;	i++){
		if(fscanf(f,	"%*s")==EOF)	return	false;
		fast	p;
		for(map<string,	Group>::iterator	mi=group.begin();	mi!=group.end();	++mi){
			if(fscanf(f,	"%f",	&p)==EOF)	return	false;
			mi->second.qua[i]=i>2?1-0.01*p:0;
		}
	}
	/*
	for(map<string,	Group>::iterator	mi=group.begin();	mi!=group.end();	++mi){
		fast	*q=mi->second.qua;
//		uint	minq=1;
//		for(uint	i=2;	i<=10;	i++)	if(q[i]<q[minq])	minq=i;
//		for(uint	i=0;	i<minq;	i++)	q[i]=0;
		q[0]=0;
		for(uint	i=1;	i<94;	i++){
			if(q[i]<0.9)	q[i]=0;
			else	q[i]=1;
		}
	}
	*/	
	null=false;	
	fclose(f);
	return	true;
}

