#include	"impute.cpp"
#include	<iostream>
#include	<cstring>
#include	<fstream>
#include	<sstream>
#include	<zlib.h>
#include	<map>
#define	Gender_Female	0
#define	Gender_Male	1

struct	SNP{
	string	chr;
	uint	pos;
	char	all[4];
	vector<fast>	prob;
};

class	SNPCall{
private:
	uint	cor,	flk,	burn,	mcmc;
	string	kno,	out;
	fast	den;
	
	uint	in,	mn;
	vector<string>	name;
	vector<SNP>	data;
	vector<char>	gender;
	vector<fast>	prob;
	map<uint,	uint>	mmap;
	map<string,	uint>	imap;
	
	void	document(void);
	bool	load(const	char	*F);	
	void	call(void);
	void	save(const	char	*F);
	void	fill_gender(const	char	*F);
	bool	load_known(const	char	*F);
public:
	void	options(int	ac,	char	**av);
};

bool	SNPCall::load(const	char	*F){
	gzFile	f=gzopen(F,	"rt");
	if(f==NULL)	return	false;
	char	*buffer=new	char[1024*1024];
	{
		gzgets(f,	buffer,	1024*1024);
		string	s;	istringstream	si(buffer);
		for(si>>s>>s>>s>>s;	!si.eof();	si>>s)	name.push_back(s);
		in=name.size();
	}
	SNP	temp;	temp.prob.resize(2*in);
	while(gzgets(f,	buffer,	1024*1024)!=NULL){
		istringstream	si(buffer);
		si>>temp.chr>>temp.pos>>temp.all;
		for(uint	i=0;	i<2*in;	i++)	si>>temp.prob[i];
		data.push_back(temp);
	}
	mn=data.size();
	delete	[]	buffer;
	gzclose(f);
	
	for(uint	m=0;	m<mn;	m++)	mmap[data[m].pos]=m;
	for(uint	i=0;	i<in;	i++)	imap[name[i]]=i;
	return	true;
}

void	SNPCall::call(void){
	uint	cn=mn%cor?mn/cor+1:mn/cor;
	real	delta=(real)mn/cn;
	prob.resize(2*in*mn);
	for(uint	c=0;	c<cn;	c++){
		uint	s=(uint)(c*delta+0.5),	e=(uint)((c+1)*delta+0.5);
		uint	bs=s>flk?s-flk:0,	be=e+flk<=mn?e+flk:mn;
		Impute	impute;
		impute.in=in;	impute.mn=be-bs;	impute.bn=burn;	impute.sn=mcmc;
		impute.prob.resize(impute.in*impute.mn*2);	impute.posi.resize(impute.mn);
		for(uint	m=0;	m<impute.mn;	m++){
			fast	*p=&impute.prob[m*impute.in*2],	*q=&data[m+bs].prob[0];
			impute.posi[m]=data[m+bs].pos;
			for(uint	i=0;	i<in;	i++,	p+=2,	q+=2){
				p[0]=q[0];	p[1]=q[1];
			}
		}
		impute.initialize(den);	impute.estimate();
		memcpy(&prob[s*in*2],	&impute.prob[(s-bs)*in*2],	(e-s)*in*2*sizeof(fast));
	}
}

void	SNPCall::save(const	char	*F){
	fast	pmin=3e-3;
	FILE	*f=fopen(F,	"wt");
	fprintf(f,	"##fileformat=VCFv4.0\n");
	fprintf(f,	"##source=BCM:SNPTools:snpcall\n");
	fprintf(f,	"##reference=1000Genomes-NCBI37\n");
	fprintf(f,	"##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n");
	fprintf(f,	"##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"Genotype Quality\">\n");	
	fprintf(f,	"#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT");
	for(uint	i=0;	i<in;	i++)	fprintf(f,	"\t%s",	name[i].c_str());
	for(uint	m=0;	m<mn;	m++){
		bool	type=isupper(data[m].all[0]);
		char	a=toupper(data[m].all[0]),	b=toupper(data[m].all[1]);
		fprintf(f,	"\n%s\t%u\t.\t",	data[m].chr.c_str(),	data[m].pos);
		fprintf(f,	"%c\t%c\t",	type?a:b,	type?b:a);
		fprintf(f,	"100\tPASS\t.\tGT:GQ");
		fast	*p=&prob[m*in*2];
		for(uint	i=0;	i<in;	i++,	p+=2){
			fast	prr=(1-p[0])*(1-p[1]),	pra=(1-p[0])*p[1]+p[0]*(1-p[1]),	paa=p[0]*p[1];
			if(prr>=pra&&prr>paa)	fprintf(f,	"\t%c/%c:%d",	type?'0':'1',	type?'0':'1',	(int)(-10*log10(1-prr+pmin)+0.5));
			else	if(pra>=prr&&pra>=paa)	fprintf(f,	"\t%c/%c:%d",	'0',	'1',	(int)(-10*log10(1-pra+pmin)+0.5));
			else	fprintf(f,	"\t%c/%c:%d",	type?'1':'0',	type?'1':'0',	(int)(-10*log10(1-paa+pmin)+0.5));
		}
	}
	fprintf(f,	"\n");
	fclose(f);
}

void	SNPCall::document(void){
	fprintf(stderr,	"\nname	snpcall 0.1");
	fprintf(stderr,	"\nfunc	genotype calling by artifical evolution");
	fprintf(stderr,	"\nauthor	Yi Wang @ Fuli Yu' Group @ BCM");
	fprintf(stderr,	"\nusage	SNPCall <data.bin gender.txt> [options]");
	fprintf(stderr,	"\n	-d <density>	relative SNP density to Sanger sequencing (0.33)");	
	fprintf(stderr,	"\n	-c <chunk>	specify number of SNPs in the core chunk (256)");
	fprintf(stderr,	"\n	-f <flank>	specify number of SNPs in each side (128)");
	fprintf(stderr,	"\n	-b <burn>	burn-in generations (64)");		
	fprintf(stderr,	"\n	-m <mcmc>	sampling generations (192)");
	fprintf(stderr,	"\n	-k <known.txt>	assign known genotypes");
	fprintf(stderr,	"\n\n");
	exit(0);
}

void	SNPCall::fill_gender(const	char	*F){
	gender.resize(in);
	for(uint	i=0;	i<in;	i++)	gender[i]=2;
	ifstream	fi(F);
	if(!fi){	cerr<<"can't open "<<F<<endl;	exit(0);	}
	char	buffer[1024];
	for(fi.getline(buffer,	1024);	!fi.eof();	fi.getline(buffer,	1024)){
		istringstream	si(buffer);
		string	temp;	si>>temp;
		for(uint	i=0;	i<in;	i++)	if(name[i]==temp){
			if(strstr(buffer,	"female")!=NULL)	gender[i]=Gender_Female;
			else	if(strstr(buffer,	"male")!=NULL)	gender[i]=Gender_Male;
		}
	}
	fi.close();
	for(uint	i=0;	i<in;	i++)	if(gender[i]==2){
		gender[i]=Gender_Female;
		cerr<<"unknown gender:\t"<<name[i]<<endl;
	}
}

bool	SNPCall::load_known(const	char	*F){
	vector<uint>	k2i;
	char	buffer[65536];
	string	temp;
	
	ifstream	fi(F);
	if(!fi)	return	false;
	fi.getline(buffer,	65536);
	istringstream	sh(buffer);
	for(sh>>temp>>temp;	!sh.eof();	sh>>temp){
		map<string,uint>::iterator	mi=imap.find(temp);
		if(mi!=imap.end())	k2i.push_back(mi->second);
		else	k2i.push_back(0xffffffff);
	}	
	uint	kn=k2i.size(),	pos;
	for(fi.getline(buffer,	65536);	!fi.eof();	fi.getline(buffer,	65536)){
		istringstream	si(buffer);
		si>>pos;	map<uint,	uint>::iterator	mi=mmap.find(pos);
		if(mi==mmap.end())	continue;
		uint	m=mi->second;	char	g[3];
		char	ref=toupper(data[m].all[0]),	alt=toupper(data[m].all[1]);
		for(uint	i=0;	i<kn;	i++){
			si>>g;	if(k2i[i]==0xffffffff)	continue;
			fast	*p=&data[m].prob[k2i[i]*2];
			char	a=toupper(g[0]),	b=toupper(g[1]);
			fast	aa=1-p[0]-p[1],	ab=p[0],	bb=p[1];
			if(a==ref&&b==ref){	aa*=0.99;	ab*=0.01;	bb*=0.001;	}
			else	if((a==ref&&b==alt)||(a==alt&&b==ref)){	aa*=0.005;	ab*=0.99;	bb*=0.005;	}
			else	if(a==alt&&b==alt){	aa*=0.001;	ab*=0.01;	bb*=0.99;	}
			p[0]=ab/(aa+ab+bb);	p[1]=bb/(aa+ab+bb);
		}
	}
	fi.close();
	return	true;
}

void	SNPCall::options(int	ac,	char	**av){
	cor=256;	flk=128;	burn=64;	mcmc=192;	den=0.33;
	char	opt;
	while((opt=getopt(ac,	av,	"d:c:f:b:m:k:"))>=0){
		switch(opt) {
		case	'd':	den=atof(optarg);	break;
		case	'c':	cor=atoi(optarg);	break;		
		case	'f':	flk=atoi(optarg);	break;
		case	'b':	burn=atoi(optarg);	break;
		case	'm':	mcmc=atoi(optarg);	break;
		case	'k':	kno=optarg;	break;
		default:	document();
		}
	}
	if(ac<optind+2)	document();
	load(av[optind]);	
	out=av[optind];	out+=".vcf";
	fill_gender(av[optind+1]);
	if(kno.size())	load_known(kno.c_str());
	call();
	save(out.c_str());
}
	
int	main(int	ac,	char	**av){
	SNPCall	sc;
	sc.options(ac,	av);
	return	0;
}

